@@ -1,4 +1,4 @@
-__version__ = "0.4.3"
+__version__ = "0.6.0"
 
 from ._graph_util import index_add, radius_graph
 from ._rotation import (
@@ -31,13 +31,22 @@ from ._rotation import (
     angles_to_xyz,
     xyz_to_angles,
 )
-from ._wigner import wigner_3j, wigner_D, wigner_J, wigner_3j_sympy, wigner_generator_x, wigner_generator_y, wigner_generator_z, wigner_rot90_y, wigner_rot_y
-from ._irreps import Irrep, Irreps, IrrepsData
+from ._su2 import su2_clebsch_gordan, su2_generators
+from ._so3 import clebsch_gordan, wigner_D, generators
+from ._instruction import Instruction
+from ._irreps import Irrep, MulIrrep, Irreps, IrrepsData
 from ._spherical_harmonics import spherical_harmonics
 from ._soft_one_hot_linspace import sus, soft_one_hot_linspace
 from ._linear import FunctionalLinear, Linear
 from ._core_tensor_product import FunctionalTensorProduct
-from ._tensor_products import FunctionalFullyConnectedTensorProduct, FullyConnectedTensorProduct, full_tensor_product, elementwise_tensor_product, FunctionalTensorSquare, TensorSquare
+from ._tensor_products import (
+    FunctionalFullyConnectedTensorProduct,
+    FullyConnectedTensorProduct,
+    full_tensor_product,
+    elementwise_tensor_product,
+    FunctionalTensorSquare,
+    TensorSquare,
+)
 from ._activation import scalar_activation, normalize_function
 from ._gate import gate
 from ._batchnorm import BatchNorm
@@ -45,12 +54,14 @@ from ._dropout import Dropout
 from ._nn import MultiLayerPerceptron
 
 __all__ = [
-    "index_add", "radius_graph",
+    "index_add",
+    "radius_graph",
     "rand_matrix",
     "identity_angles",
     "rand_angles",
     "compose_angles",
     "inverse_angles",
+    "Instruction",
     "identity_quaternion",
     "rand_quaternion",
     "compose_quaternion",
@@ -74,14 +85,29 @@ __all__ = [
     "axis_angle_to_angles",
     "angles_to_xyz",
     "xyz_to_angles",
-    "wigner_3j", "wigner_3j_sympy", "wigner_D", "wigner_J", "wigner_generator_x", "wigner_generator_y", "wigner_generator_z", "wigner_rot90_y", "wigner_rot_y",
-    "Irrep", "Irreps", "IrrepsData",
+    "su2_clebsch_gordan",
+    "su2_generators",
+    "clebsch_gordan",
+    "wigner_D",
+    "generators",
+    "Irrep",
+    "MulIrrep",
+    "Irreps",
+    "IrrepsData",
     "spherical_harmonics",
-    "sus", "soft_one_hot_linspace",
-    "FunctionalLinear", "Linear",
+    "sus",
+    "soft_one_hot_linspace",
+    "FunctionalLinear",
+    "Linear",
     "FunctionalTensorProduct",
-    "FunctionalFullyConnectedTensorProduct", "FullyConnectedTensorProduct", "full_tensor_product", "elementwise_tensor_product", "FunctionalTensorSquare", "TensorSquare",
-    "scalar_activation", "normalize_function",
+    "FunctionalFullyConnectedTensorProduct",
+    "FullyConnectedTensorProduct",
+    "full_tensor_product",
+    "elementwise_tensor_product",
+    "FunctionalTensorSquare",
+    "TensorSquare",
+    "scalar_activation",
+    "normalize_function",
     "gate",
     "BatchNorm",
     "Dropout",
