@@ -11,13 +11,15 @@ def normalize_function(phi):
     with jax.ensure_compile_time_eval():
         k = jax.random.PRNGKey(0)
         x = jax.random.normal(k, (1_000_000,))
-        c = jnp.mean(phi(x)**2)**0.5
+        c = jnp.mean(phi(x) ** 2) ** 0.5
 
         if jnp.allclose(c, 1.0):
             return phi
         else:
+
             def rho(x):
                 return phi(x) / c
+
             return rho
 
 
@@ -51,13 +53,17 @@ def scalar_activation(input: IrrepsData, acts: List[Optional[Callable[[float], f
     for (mul, (l_in, p_in)), x, act in zip(input.irreps, input.list, acts):
         if act is not None:
             if l_in != 0:
-                raise ValueError(f"Activation: cannot apply an activation function to a non-scalar input. {input.irreps} {acts}")
+                raise ValueError(
+                    f"Activation: cannot apply an activation function to a non-scalar input. {input.irreps} {acts}"
+                )
 
             act = normalize_function(act)
 
             p_out = parity_function(act) if p_in == -1 else p_in
             if p_out == 0:
-                raise ValueError("Activation: the parity is violated! The input scalar is odd but the activation is neither even nor odd.")
+                raise ValueError(
+                    "Activation: the parity is violated! The input scalar is odd but the activation is neither even nor odd."
+                )
 
             irreps_out.append((mul, (0, p_out)))
             if x is None:
@@ -76,7 +82,7 @@ def scalar_activation(input: IrrepsData, acts: List[Optional[Callable[[float], f
     if acts and acts.count(acts[0]) == len(acts):
         # for performance, if all the activation functions are the same, we can apply it to the contiguous array as well
         contiguous = input.contiguous if acts[0] is None else normalize_function(acts[0])(input.contiguous)
-        return IrrepsData(irreps_out, contiguous, list)
+        return IrrepsData(irreps=irreps_out, contiguous=contiguous, list=list)
 
     return IrrepsData.from_list(irreps_out, list, input.shape)
 
@@ -103,7 +109,7 @@ def key_value_activation(phi, key, value):
     assert value.ndim == 1
 
     d = value.shape[0]
-    key = key / jnp.sqrt(1/16 + jnp.sum(key**2))  # 1/16 is arbitrary small... but not too small...
+    key = key / jnp.sqrt(1 / 16 + jnp.sum(key ** 2))  # 1/16 is arbitrary small... but not too small...
     scalar = jnp.sum(key * value)
     scalar = normalize_function(phi)(scalar)
-    return d**0.5 * scalar * key  # component normalized
+    return d ** 0.5 * scalar * key  # component normalized
