@@ -4,20 +4,21 @@ import math
 from functools import partial
 from typing import List, Optional, Tuple
 
+import dataclasses
 import jax
 import jax.numpy as jnp
 import jax.scipy
 
 from e3nn_jax import matrix_to_angles, perm, quaternion_to_angles, wigner_D
 
-from .dataclasses import dataclass, static_field
 
-
-class Irrep(tuple):
+@dataclasses.dataclass(init=False, frozen=True)
+class Irrep:
     r"""Irreducible representation of :math:`O(3)`
 
     This class does not contain any data, it is a structure that describe the representation.
-    It is typically used as argument of other classes of the library to define the input and output representations of functions.
+    It is typically used as argument of other classes of the library to define the input and output
+    representations of functions.
 
     Args:
         l: non-negative integer, the degree of the representation, :math:`l = 0, 1, \dots`
@@ -48,10 +49,14 @@ class Irrep(tuple):
         >>> Irrep("1o") + Irrep("2o")
         1x1o+1x2o
     """
-    def __new__(cls, l, p=None):
+    l: int
+    p: int
+
+    def __init__(self, l, p=None):
         if p is None:
             if isinstance(l, Irrep):
-                return l
+                p = l.p
+                l = l.l
 
             if isinstance(l, str):
                 try:
@@ -59,31 +64,22 @@ class Irrep(tuple):
                     l = int(name[:-1])
                     assert l >= 0
                     p = {
-                        'e': 1,
-                        'o': -1,
-                        'y': (-1)**l,
+                        "e": 1,
+                        "o": -1,
+                        "y": (-1) ** l,
                     }[name[-1]]
                 except Exception:
-                    raise ValueError(f"unable to convert string \"{name}\" into an Irrep")
+                    raise ValueError(f'unable to convert string "{name}" into an Irrep')
             elif isinstance(l, tuple):
                 l, p = l
 
         assert isinstance(l, int) and l >= 0, l
         assert p in [-1, 1], p
-        return super().__new__(cls, (l, p))
-
-    @property
-    def l(self) -> int:
-        r"""The degree of the representation, :math:`l = 0, 1, \dots`."""
-        return self[0]
-
-    @property
-    def p(self) -> int:
-        r"""The parity of the representation, :math:`p = \pm 1`."""
-        return self[1]
+        object.__setattr__(self, "l", l)
+        object.__setattr__(self, "p", p)
 
     def __repr__(self):
-        p = {+1: 'e', -1: 'o'}[self.p]
+        p = {+1: "e", -1: "o"}[self.p]
         return f"{self.l}{p}"
 
     @classmethod
@@ -96,8 +92,8 @@ class Irrep(tuple):
             (0e, 0o, 1o, 1e)
         """
         for l in itertools.count():
-            yield Irrep(l, (-1)**l)
-            yield Irrep(l, -(-1)**l)
+            yield Irrep(l, (-1) ** l)
+            yield Irrep(l, -((-1) ** l))
 
             if l == lmax:
                 break
@@ -126,7 +122,7 @@ class Irrep(tuple):
         """
         alpha, beta, gamma, k = jnp.broadcast_arrays(alpha, beta, gamma, k)
         k = jnp.asarray(k)
-        return wigner_D(self.l, alpha, beta, gamma) * self.p**k[..., None, None]
+        return wigner_D(self.l, alpha, beta, gamma) * self.p ** k[..., None, None]
 
     def D_from_quaternion(self, q, k=0):
         r"""Matrix of the representation, see `Irrep.D_from_angles`
@@ -187,12 +183,6 @@ class Irrep(tuple):
         for l in range(lmin, lmax + 1):
             yield Irrep(l, p)
 
-    def count(self, _value):
-        raise NotImplementedError
-
-    def index(self, _value):
-        raise NotImplementedError
-
     def __rmul__(self, other):
         r"""
         >>> 3 * Irrep('1e')
@@ -204,29 +194,28 @@ class Irrep(tuple):
     def __add__(self, other):
         return Irreps(self) + Irreps(other)
 
-    def __contains__(self, _object):
-        raise NotImplementedError
+    def __iter__(self):
+        yield self.l
+        yield self.p
 
-    def __len__(self):
-        raise NotImplementedError
+    def __lt__(self, other):
+        return (self.l, self.p) < (other.l, other.p)
 
 
-class _MulIr(tuple):
-    def __new__(cls, mul, ir=None):
+jax.tree_util.register_pytree_node(Irrep, lambda ir: ((), ir), lambda ir, _: ir)
+
+
+@dataclasses.dataclass(init=False, frozen=True)
+class MulIrrep:
+    mul: int
+    ir: Irrep
+
+    def __init__(self, mul, ir=None):
         if ir is None:
             mul, ir = mul
 
-        assert isinstance(mul, int)
-        assert isinstance(ir, Irrep)
-        return super().__new__(cls, (mul, ir))
-
-    @property
-    def mul(self) -> int:
-        return self[0]
-
-    @property
-    def ir(self) -> Irrep:
-        return self[1]
+        object.__setattr__(self, "mul", mul)
+        object.__setattr__(self, "ir", ir)
 
     @property
     def dim(self) -> int:
@@ -235,18 +224,23 @@ class _MulIr(tuple):
     def __repr__(self):
         return f"{self.mul}x{self.ir}"
 
-    def count(self, _value):
-        raise NotImplementedError
+    def __iter__(self):
+        yield self.mul
+        yield self.ir
 
-    def index(self, _value):
-        raise NotImplementedError
+    def __lt__(self, other):
+        return (self.mul, self.ir) < (other.mul, other.ir)
+
+
+jax.tree_util.register_pytree_node(MulIrrep, lambda mulir: ((), mulir), lambda mulir, _: mulir)
 
 
 class Irreps(tuple):
     r"""Direct sum of irreducible representations of :math:`O(3)`
 
     This class does not contain any data, it is a structure that describe the representation.
-    It is typically used as argument of other classes of the library to define the input and output representations of functions.
+    It is typically used as argument of other classes of the library to define the input and output
+    representations of functions.
 
     Attributes:
         dim (int): the total dimension of the representation
@@ -279,19 +273,20 @@ class Irreps(tuple):
         >>> Irreps(), Irreps("")
         (, )
     """
+
     def __new__(cls, irreps=None):
         if isinstance(irreps, Irreps):
             return super().__new__(cls, irreps)
 
         out = []
         if isinstance(irreps, Irrep):
-            out.append(_MulIr(1, Irrep(irreps)))
+            out.append(MulIrrep(1, Irrep(irreps)))
         elif isinstance(irreps, str):
             try:
                 if irreps.strip() != "":
-                    for mul_ir in irreps.split('+'):
-                        if 'x' in mul_ir:
-                            mul, ir = mul_ir.split('x')
+                    for mul_ir in irreps.split("+"):
+                        if "x" in mul_ir:
+                            mul, ir = mul_ir.split("x")
                             mul = int(mul)
                             ir = Irrep(ir)
                         else:
@@ -299,9 +294,9 @@ class Irreps(tuple):
                             ir = Irrep(mul_ir)
 
                         assert isinstance(mul, int) and mul >= 0
-                        out.append(_MulIr(mul, ir))
+                        out.append(MulIrrep(mul, ir))
             except Exception:
-                raise ValueError(f"Unable to convert string \"{irreps}\" into an Irreps")
+                raise ValueError(f'Unable to convert string "{irreps}" into an Irreps')
         elif irreps is None:
             pass
         else:
@@ -312,7 +307,7 @@ class Irreps(tuple):
                 elif isinstance(mul_ir, Irrep):
                     mul = 1
                     ir = mul_ir
-                elif isinstance(mul_ir, _MulIr):
+                elif isinstance(mul_ir, MulIrrep):
                     mul, ir = mul_ir
                 elif isinstance(mul_ir, int):
                     mul, ir = 1, Irrep(l=mul_ir, p=1)
@@ -324,9 +319,9 @@ class Irreps(tuple):
                     ir = None
 
                 if not (isinstance(mul, int) and mul >= 0 and ir is not None):
-                    raise ValueError(f"Unable to interpret \"{mul_ir}\" as an irrep.")
+                    raise ValueError(f'Unable to interpret "{mul_ir}" as an irrep.')
 
-                out.append(_MulIr(mul, ir))
+                out.append(MulIrrep(mul, ir))
         return super().__new__(cls, out)
 
     @staticmethod
@@ -346,7 +341,7 @@ class Irreps(tuple):
             >>> Irreps.spherical_harmonics(4, p=1)
             1x0e+1x1e+1x2e+1x3e+1x4e
         """
-        return Irreps([(1, (l, p**l)) for l in range(lmax + 1)])
+        return Irreps([(1, (l, p ** l)) for l in range(lmax + 1)])
 
     def slices(self):
         r"""List of slices corresponding to indices for each irrep.
@@ -362,7 +357,7 @@ class Irreps(tuple):
             i += mul_ir.dim
         return s
 
-    def randn(self, key, size, *, normalization='component'):
+    def randn(self, key, size, *, normalization="component"):
         r"""Random tensor.
 
         Args:
@@ -383,11 +378,11 @@ class Irreps(tuple):
         """
         di = size.index(-1)
         lsize = size[:di]
-        rsize = size[di + 1:]
+        rsize = size[di + 1 :]
 
-        if normalization == 'component':
+        if normalization == "component":
             return jax.random.normal(key, lsize + (self.dim,) + rsize)
-        elif normalization == 'norm':
+        elif normalization == "norm":
             x = jnp.zeros(lsize + (self.dim,) + rsize)
             for s, (mul, ir) in zip(self.slices(), self):
                 key, k = jax.random.split(key)
@@ -661,23 +656,49 @@ class Irreps(tuple):
         return self.D_from_angles(*matrix_to_angles(R), k)
 
 
-@dataclass
+jax.tree_util.register_pytree_node(Irreps, lambda irreps: ((), irreps), lambda irreps, _: irreps)
+
+
 class IrrepsData:
     r"""Class storing data and its irreps"""
 
-    irreps: Irreps = static_field()
-    contiguous: jnp.array
-    list: List[Optional[jnp.array]]
+    irreps: Irreps
+    contiguous: jnp.ndarray  # this field is mendatory because it contains the shape
+    _list: List[Optional[jnp.ndarray]]  # this field is lazy, it is computed only when needed
+
+    def __init__(self, irreps: Irreps, contiguous: jnp.ndarray, list: List[Optional[jnp.ndarray]] = None):
+        self.irreps = Irreps(irreps)
+        self.contiguous = contiguous
+        self._list = list
+
+    @property
+    def list(self) -> List[Optional[jnp.ndarray]]:
+        if self._list is None:
+            shape = self.contiguous.shape[:-1]
+            if len(self.irreps) == 1:
+                mul, ir = self.irreps[0]
+                list = [jnp.reshape(self.contiguous, shape + (mul, ir.dim))]
+            else:
+                list = [
+                    jnp.reshape(self.contiguous[..., i], shape + (mul, ir.dim))
+                    for i, (mul, ir) in zip(self.irreps.slices(), self.irreps)
+                ]
+            self._list = list
+        return self._list
 
     @staticmethod
     def zeros(irreps: Irreps, shape) -> "IrrepsData":
         irreps = Irreps(irreps)
-        return IrrepsData(irreps, jnp.zeros(shape + (irreps.dim,)), [None] * len(irreps))
+        return IrrepsData(irreps=irreps, contiguous=jnp.zeros(shape + (irreps.dim,)), list=[None] * len(irreps))
 
     @staticmethod
     def ones(irreps: Irreps, shape) -> "IrrepsData":
         irreps = Irreps(irreps)
-        return IrrepsData(irreps, jnp.ones(shape + (irreps.dim,)), [jnp.ones(shape + (mul, ir.dim)) for mul, ir in irreps])
+        return IrrepsData(
+            irreps=irreps,
+            contiguous=jnp.ones(shape + (irreps.dim,)),
+            list=[jnp.ones(shape + (mul, ir.dim)) for mul, ir in irreps],
+        )
 
     @staticmethod
     def new(irreps: Irreps, any) -> "IrrepsData":
@@ -716,21 +737,22 @@ class IrrepsData:
         irreps = Irreps(irreps)
         assert len(irreps) == len(list), f"irreps has {len(irreps)} elements and list has {len(list)}"
         assert all(x is None or isinstance(x, jnp.ndarray) for x in list)
-        assert all(
-            x is None or x.shape == shape + (mul, ir.dim)
-            for x, (mul, ir) in zip(list, irreps)
-        ), ([x.shape for x in list if x is not None], shape)
+        assert all(x is None or x.shape == shape + (mul, ir.dim) for x, (mul, ir) in zip(list, irreps)), (
+            [x.shape for x in list if x is not None],
+            shape,
+        )
 
         if irreps.dim > 0:
-            contiguous = jnp.concatenate([
-                jnp.zeros(shape + (mul_ir.dim,))
-                if x is None else
-                x.reshape(shape + (mul_ir.dim,))
-                for mul_ir, x in zip(irreps, list)
-            ], axis=-1)
+            contiguous = jnp.concatenate(
+                [
+                    jnp.zeros(shape + (mul_ir.dim,)) if x is None else x.reshape(shape + (mul_ir.dim,))
+                    for mul_ir, x in zip(irreps, list)
+                ],
+                axis=-1,
+            )
         else:
             contiguous = jnp.zeros(shape + (0,))
-        return IrrepsData(irreps, contiguous, list)
+        return IrrepsData(irreps=irreps, contiguous=contiguous, list=list)
 
     @staticmethod
     def from_contiguous(irreps: Irreps, contiguous) -> "IrrepsData":
@@ -743,19 +765,8 @@ class IrrepsData:
         Returns:
             IrrepsData
         """
-        irreps = Irreps(irreps)
-        assert contiguous.shape[-1] == irreps.dim
-
-        shape = contiguous.shape[:-1]
-        if len(irreps) == 1:
-            mul, ir = irreps[0]
-            list = [jnp.reshape(contiguous, shape + (mul, ir.dim))]
-        else:
-            list = [
-                jnp.reshape(contiguous[..., i], shape + (mul, ir.dim))
-                for i, (mul, ir) in zip(irreps.slices(), irreps)
-            ]
-        return IrrepsData(irreps, contiguous, list)
+        assert contiguous.shape[-1] == Irreps(irreps).dim
+        return IrrepsData(irreps=irreps, contiguous=contiguous, list=None)
 
     def __repr__(self):
         return f"IrrepsData({self.irreps}, {self.contiguous}, {self.list})"
@@ -769,27 +780,20 @@ class IrrepsData:
         return len(self.shape)
 
     def reshape(self, shape) -> "IrrepsData":
-        list = [
-            None if x is None else x.reshape(shape + (mul, ir.dim))
-            for (mul, ir), x in zip(self.irreps, self.list)
-        ]
-        return IrrepsData(self.irreps, self.contiguous.reshape(shape + (self.irreps.dim,)), list)
+        list = [None if x is None else x.reshape(shape + (mul, ir.dim)) for (mul, ir), x in zip(self.irreps, self.list)]
+        return IrrepsData(irreps=self.irreps, contiguous=self.contiguous.reshape(shape + (self.irreps.dim,)), list=list)
 
     def broadcast_to(self, shape) -> "IrrepsData":
         assert isinstance(shape, tuple)
         contiguous = jnp.broadcast_to(self.contiguous, shape + (self.irreps.dim,))
         list = [
-            None if x is None else jnp.broadcast_to(x, shape + (mul, ir.dim))
-            for (mul, ir), x in zip(self.irreps, self.list)
+            None if x is None else jnp.broadcast_to(x, shape + (mul, ir.dim)) for (mul, ir), x in zip(self.irreps, self.list)
         ]
-        return IrrepsData(self.irreps, contiguous, list)
+        return IrrepsData(irreps=self.irreps, contiguous=contiguous, list=list)
 
     def replace_none_with_zeros(self) -> "IrrepsData":
-        list = [
-            jnp.zeros(self.shape + (mul, ir.dim)) if x is None else x
-            for (mul, ir), x in zip(self.irreps, self.list)
-        ]
-        return IrrepsData(self.irreps, self.contiguous, list)
+        list = [jnp.zeros(self.shape + (mul, ir.dim)) if x is None else x for (mul, ir), x in zip(self.irreps, self.list)]
+        return IrrepsData(irreps=self.irreps, contiguous=self.contiguous, list=list)
 
     def remove_nones(self) -> "IrrepsData":
         if any(x is None for x in self.list):
@@ -805,7 +809,7 @@ class IrrepsData:
         contiguous_parts = jnp.split(self.contiguous, [self.irreps[:i].dim for i in indices], axis=-1)
         assert len(contiguous_parts) == len(indices) + 1
         return [
-            IrrepsData(self.irreps[i:j], contiguous, self.list[i:j])
+            IrrepsData(irreps=self.irreps[i:j], contiguous=contiguous, list=self.list[i:j])
             for (i, j), contiguous in zip(zip([0] + indices, indices + [len(self.irreps)]), contiguous_parts)
         ]
 
@@ -818,10 +822,7 @@ class IrrepsData:
     def repeat_mul_by_last_axis(self) -> "IrrepsData":
         assert len(self.shape) >= 1
         irreps = Irreps([(self.shape[-1] * mul, ir) for mul, ir in self.irreps])
-        list = [
-            None if x is None else x.reshape(self.shape[:-1] + (mul, ir.dim))
-            for (mul, ir), x in zip(irreps, self.list)
-        ]
+        list = [None if x is None else x.reshape(self.shape[:-1] + (mul, ir.dim)) for (mul, ir), x in zip(irreps, self.list)]
         return IrrepsData.from_list(irreps, list, self.shape[:-1])
 
     def factor_irreps_to_last_axis(self) -> "IrrepsData":
@@ -836,13 +837,11 @@ class IrrepsData:
 
         irreps = Irreps([(mul // factor, ir) for mul, ir in self.irreps])
         list = [
-            None if x is None else x.reshape(self.shape + (factor, mul, ir.dim))
-            for (mul, ir), x in zip(irreps, self.list)
+            None if x is None else x.reshape(self.shape + (factor, mul, ir.dim)) for (mul, ir), x in zip(irreps, self.list)
         ]
         return IrrepsData.from_list(irreps, list, self.shape + (factor,))
 
-    @partial(jax.jit, inline=True)
-    def transform_by_angles(self, alpha, beta, gamma, k=0):
+    def transform_by_angles(self, alpha: float, beta: float, gamma: float, k: int = 0) -> "IrrepsData":
         r"""Rotate the data by angles according to the irreps
 
         Args:
@@ -857,14 +856,12 @@ class IrrepsData:
         # Optimization: we use only the list of arrays, not the contiguous data
         D = {ir: ir.D_from_angles(alpha, beta, gamma, k) for ir in {ir for _, ir in self.irreps}}
         new_list = [
-            jnp.reshape(jnp.einsum("ij,...uj->...ui", D[ir], x), self.shape + (mul, ir.dim))
-            if x is not None else None
+            jnp.reshape(jnp.einsum("ij,...uj->...ui", D[ir], x), self.shape + (mul, ir.dim)) if x is not None else None
             for (mul, ir), x in zip(self.irreps, self.list)
         ]
         return IrrepsData.from_list(self.irreps, new_list, self.shape)
 
-    @partial(jax.jit, inline=True)
-    def transform_by_quaternion(self, q, k=0):
+    def transform_by_quaternion(self, q: jnp.ndarray, k: int = 0) -> "IrrepsData":
         r"""Rotate data by a rotation given by a quaternion
 
         Args:
@@ -876,8 +873,7 @@ class IrrepsData:
         """
         return self.transform_by_angles(*quaternion_to_angles(q), k)
 
-    @partial(jax.jit, inline=True)
-    def transform_by_matrix(self, R):
+    def transform_by_matrix(self, R: jnp.ndarray) -> "IrrepsData":
         r"""Rotate data by a rotation given by a matrix
 
         Args:
@@ -974,27 +970,24 @@ class IrrepsData:
 
         assert len(new_list) == len(irreps)
         assert all(x is None or isinstance(x, jnp.ndarray) for x in new_list)
-        assert all(
-            x is None or x.shape[-2:] == (mul, ir.dim)
-            for x, (mul, ir) in zip(new_list, irreps)
-        )
+        assert all(x is None or x.shape[-2:] == (mul, ir.dim) for x, (mul, ir) in zip(new_list, irreps))
 
-        return IrrepsData(irreps, self.contiguous, new_list)
+        return IrrepsData(irreps=irreps, contiguous=self.contiguous, list=new_list)
 
     def __add__(self, other: "IrrepsData") -> "IrrepsData":
         assert self.irreps == other.irreps
         list = [x if y is None else (y if x is None else x + y) for x, y in zip(self.list, other.list)]
-        return IrrepsData(self.irreps, self.contiguous + other.contiguous, list)
+        return IrrepsData(irreps=self.irreps, contiguous=self.contiguous + other.contiguous, list=list)
 
     def __sub__(self, other: "IrrepsData") -> "IrrepsData":
         assert self.irreps == other.irreps
         list = [x if y is None else (-y if x is None else x - y) for x, y in zip(self.list, other.list)]
-        return IrrepsData(self.irreps, self.contiguous - other.contiguous, list)
+        return IrrepsData(irreps=self.irreps, contiguous=self.contiguous - other.contiguous, list=list)
 
     def __mul__(self, other) -> "IrrepsData":
         other = jnp.array(other)
         list = [None if x is None else x * other[..., None, None] for x in self.list]
-        return IrrepsData(self.irreps, self.contiguous * other[..., None], list)
+        return IrrepsData(irreps=self.irreps, contiguous=self.contiguous * other[..., None], list=list)
 
     def __rmul__(self, other) -> "IrrepsData":
         return self * other
@@ -1002,7 +995,7 @@ class IrrepsData:
     def __truediv__(self, other) -> "IrrepsData":
         other = jnp.array(other)
         list = [None if x is None else x / other[..., None, None] for x in self.list]
-        return IrrepsData(self.irreps, self.contiguous / other[..., None], list)
+        return IrrepsData(irreps=self.irreps, contiguous=self.contiguous / other[..., None], list=list)
 
     @staticmethod
     def cat(args, axis="irreps"):
@@ -1019,9 +1012,9 @@ class IrrepsData:
         if axis == "irreps":
             irreps = Irreps(sum([x.irreps for x in args], Irreps("")))
             return IrrepsData(
-                irreps,
-                jnp.concatenate([x.contiguous for x in args], axis=-1),
-                sum([x.list for x in args], [])
+                irreps=irreps,
+                contiguous=jnp.concatenate([x.contiguous for x in args], axis=-1),
+                list=sum([x.list for x in args], []),
             )
         elif axis == "mul":
             raise NotImplementedError
@@ -1032,13 +1025,20 @@ class IrrepsData:
             axis += len(args[0].shape)
         args = [x.replace_none_with_zeros() for x in args]  # TODO this could be optimized
         return IrrepsData(
-            args[0].irreps,
-            jnp.concatenate([x.contiguous for x in args], axis=axis),
-            [jnp.concatenate(xs, axis=axis) for xs in zip(*[x.list for x in args])]
+            irreps=args[0].irreps,
+            contiguous=jnp.concatenate([x.contiguous for x in args], axis=axis),
+            list=[jnp.concatenate(xs, axis=axis) for xs in zip(*[x.list for x in args])],
         )
 
     @staticmethod
-    def randn(irreps, key, size=(), *, normalization='component'):
+    def randn(irreps, key, size=(), *, normalization="component"):
         irreps = Irreps(irreps)
         x = irreps.randn(key, size + (-1,), normalization=normalization)
         return IrrepsData.from_contiguous(irreps, x)
+
+
+jax.tree_util.register_pytree_node(
+    IrrepsData,
+    lambda x: ((x.contiguous, x.list), x.irreps),
+    lambda x, data: IrrepsData(irreps=x, contiguous=data[0], list=data[1]),
+)
